from cryptography.fernet import Fernet, InvalidToken


class Security:
    '''
    Cryptography for Terminal Chat Application.
    '''

    def __init__(self, key, digest_count):
        self.fernet = Fernet(key)
        self.digest_count = digest_count
    

    def encrypt(self, message):
        '''
        Encrypts the `message` based on `self.digest_count`.

        `message` can be an instance of bytes, str, or int.
        '''

        # Make the message to be an instance of bytes if 
        # its just str or int.
        if not isinstance(message, bytes):
            if isinstance(message, int):
                message = str(message)
            message = message.encode()

        for index, _ in enumerate(range(self.digest_count)):
            if index == 0:
                encrypted_message = self.fernet.encrypt(message)
                continue
            encrypted_message = self.fernet.encrypt(encrypted_message)
        return encrypted_message
    

    def decrypt(self, encrypted_message):
        '''
        Decrypts the `encrypted_message` based on `self.digest_count`.

        `encrypted_message` can be an instance of bytes or str.
        '''

        # Make the `encrypted_message` to be an instance of bytes if 
        # its just str.
        if not isinstance(encrypted_message, bytes):
            encrypted_message = encrypted_message.encode()

        try:
            for index, _ in enumerate(range(self.digest_count)):
                if index == 0:
                    decrypted_message = self.fernet.decrypt(encrypted_message)
                    continue
                decrypted_message = self.fernet.decrypt(decrypted_message)
            return decrypted_message
        except InvalidToken:
            # Return None if `encrypted_message` can't be encrypted.
            # This is to avoid using the try...except... statement
            # when using this.
            return None


# Idea from:
# https://stackoverflow.com/questions/2408560/non-blocking-console-input


import threading


class NonBlockingInput(threading.Thread):
    '''A non-blocking `input()`.'''

    def __init__(
        self, 
        callback,
        input_prefix,
        daemon=True,
        thread_name='non-blocking-input-thread'
        ):
        
        self.callback = callback
        self.prefix = input_prefix
        super(NonBlockingInput, self).__init__(name=thread_name)
        
        # If true, this dies when the main thread is dead.
        self.daemon = daemon
        
        self.start()
        

    def run(self):
        while True:
            self.callback(input(self.prefix))


import json


def create_event(type, **kwargs):
    '''
    Returns an event with type of given `type`. The event is JSON
    encoded.
    '''
    return json.dumps({'type': type, **kwargs})


def create_message_event(_from, message):
    '''
    Event for message of websocket clients.
    '''
    event = {'from': _from, 'message': message}
    return create_event('message', **event)


def create_users_event(users):
    '''
    Event for clients that is connected to the server.
    '''
    return create_event('users', users=users)


from argparse import ArgumentParser

from .client import Client
from .server import Server


class CommandLineInterface:
    '''
    Command Line Interface for Terminal Chat Application.
    '''

    def __init__(self):
        # Top level parser for client and server program.
        parser = ArgumentParser(
            prog='Terminal Chat Application',
            description=self.__doc__
        )
        
        # Subparser for client and server program.
        subparser = parser.add_subparsers(
            help='Choose a program between client and server.',
            dest='program'
        )

        # Server program parser.
        server_parser = subparser.add_parser(
            'server', 
            prog='Server Program',
            help='Run the program as server.'
        )
        server_parser.add_argument(
            '-p',
            '--port',
            default=1719,
            type=int,
            help='''
            Server port. Default is 1719. 
            '''
        )
        server_parser.add_argument(
            '--password',
            default='top_secret',
            help='''
            Password for authorization header. 
            Default is `top_secret`.
            '''
        )
        server_parser.add_argument(
            '-cdc',
            '--cryptography-digest-count',
            default=3,
            type=int,
            help='''
            Digest count for cryptography of server 
            and client program. Max of 5 and min of 1.
            ''',
            dest='digest_count'
        )

        # Client program parser.
        client_parser = subparser.add_parser(
            'client',
            prog='Client Program',
            help='Run the program as client.'
        )
        client_parser.add_argument(
            'cryptography_key',
            default='ws://localhost:1719/',
            help='''
            Cryptography key for cryptography of server 
            and client program.
            '''
        )
        client_parser.add_argument(
            '--url',
            default='ws://localhost:1719/',
            help='''
            URL where the client will connect. Default is 
            "ws://localhost:1719/".
            ''',
            dest='websocket_url'
        )
        client_parser.add_argument(
            '--username',
            default='',
            help='''
            Username for username header of client. If you don't 
            use this optional arg, the Client Program will create 
            its own username. Starting with user then number. 
            Example: "user_1234".
            '''
        )
        client_parser.add_argument(
            '--password',
            default='top_secret',
            help='''
            Password for authorization header. 
            Default is `top_secret`.
            '''
        )
        client_parser.add_argument(
            '-cdc',
            '--cryptography-digest-count',
            default=3,
            type=int,
            help='''
            Digest count for cryptography of server 
            and client program. Max of 5 and min of 1.
            ''',
            dest='digest_count'
        )
        
        self.args = parser.parse_args()
        
        # Validates if digest count for cryptography
        # is in between 1-5. If not raise
        # `parser.error()`.
        if not self.args.digest_count in range(1, 6):
            # Chooce which parser to use. If client
            # or server program.
            error_info = f'Invalid digest count for cryptography. It should be in between of 1-5 but got "{self.args.digest_count}".'
            if self.args.program == 'server':
                raise server_parser.error(error_info)
            raise client_parser.error(error_info)
    
    
    def run(self):
        '''
        Runs the program.
        '''
        try:
            # Run the program as server
            if self.args.program == 'server':
                server = Server(
                    port=self.args.port, 
                    password=self.args.password,
                    cryptography_digest_count=self.args.digest_count
                )
                server.run()

            # Run the program as client
            elif self.args.program == 'client':
                client = Client(
                    cryptography_key=self.args.cryptography_key,
                    cryptography_digest_count=self.args.digest_count,
                    url=self.args.websocket_url,
                    username=self.args.username,
                    password=self.args.password
                )
                
                client.run()
        except KeyboardInterrupt:
            if self.args.program == 'server':
                print('\nServer has been stopped.')
            elif self.args.program == 'client':
                if client._successfully_connected:
                    print(f'\nDisconnected to the `{client.url}` server.')


import asyncio
import json

import websockets
from cryptography.fernet import Fernet

from .event import (
    create_message_event, 
    create_users_event
)
from .security import Security


class Server:
    '''
    Server API for Terminal Chat Application.
    '''

    def __init__(
        self, 
        port, 
        password='top_secret', 
        cryptography_digest_count=3
        ):

        self.port = int(port)
        self.password = password
        self.loop = asyncio.get_event_loop()
        
        # For cryptography of messages of client, 
        # authorization and username header.
        self.cryptography_key = Fernet.generate_key().decode()
        self.security = Security(
            self.cryptography_key,
            cryptography_digest_count
        )

        # For clients that is connected
        self.users = set()
        self.usernames = []
    
    
    async def server(self, websocket, path):
        '''
        The main handler of the websocket request.

        This first register the websocket to the server, 
        if it successful, it will allow the websocket 
        to receive a message from any websocket 
        that is connected to the server.
        '''
        
        # Register the websocket request to server. 
        # If registering is not successful, reject 
        # the request.
        is_registered = await self.register(websocket, path)
        if not is_registered:
            await websocket.close(
                1008,
                'Invalid authorization header or username header is already registered on the server.'
            )
            
            # Show to console what remote address is rejected.
            print(
                '[Rejected]', 
                f'{websocket.remote_address[0]}:{websocket.remote_address[1]}'
            )
            return

        try:
            # When there is a message from
            # any websocket, decrypt the 
            # message, and send it to all 
            # websocket that is connected 
            # to the server.
            # If message can't be decrypted,
            # close the connection since,
            # only a valid websocket can send
            # a right encrypted message.
            async for message in websocket:
                message = self.security.decrypt(message)
                if message is None:
                    await websocket.close(
                        1008,
                        'Invalid message. Make sure it is encrypted with the same cryptography key from the server.'
                    )
                    break
                message = json.loads(message)
                if message:
                    await self.notify_all_user(
                        create_message_event(
                            message['from'],
                            message['message']
                        )
                    )
        finally:
            # Before or after the websocket disconnect to the server
            # unregister them.
            await self.unregister(websocket, path)


    async def register(self, websocket, path):
        '''
        First it checks if the websocket is authorized,
        if not, it returns False. Then it register
        the websocket to the server by adding the 
        websocket to `self.users` and its username 
        to `self.usernames`. If the websocket does 
        not have username, it will only add the 
        websocket to `self.users`. If the websocket 
        username is already registered, this will 
        return False, as we don't want a duplicate 
        username.

        It returns True if registering the websocket 
        is successful, otherwise False.
        '''

        # Check if the websocket is authorized, if not
        # return False
        if not self.is_authorized(websocket):
            return False

        # Add the websocket to `self.users` and websocket
        # username to `self.usernames` if it has.
        # If the websocket username is already registered,
        # return False.
        username = self.get_username_header(websocket)
        if username is not None:
            if username in self.usernames:
                return False
            self.usernames.append(username)
        self.users.add(websocket)
        
        
        # Notify the users how many user is connected to the
        # websocket server.
        await self.notify_all_user(create_users_event(len(self.users)))
        
        # Show to console who connect to the 
        # websocket server.
        print('[Connected]')
        print('Username:', 'No username' if username is None else username)
        print(
            'Address:', 
            f'{websocket.remote_address[0]}:{websocket.remote_address[1]}'
        )
        print('[Path]', path)

        # Successfully registered.
        return True
        

    async def unregister(self, websocket, path):
        '''
        Unregisters the websocket to the server by 
        removing the websocket to `self.users` and 
        its username to `self.usernames`. If
        the websocket does not have username,
        it will only remove the websocket to 
        `self.users`.
        '''

        # Remove the websocket to `self.users` and websocket
        # username to `self.usernames` if it has.
        username = self.get_username_header(websocket)
        if username is not None:
            self.usernames.remove(username)
        self.users.remove(websocket)

        await self.notify_all_user(create_users_event(len(self.users)))
        
        # Show to console who disconnect
        # to the websocket server.
        print('[Disconnected]')
        print('Username:', 'No username' if username is None else username)
        print(
            'Address:', 
            f'{websocket.remote_address[0]}:{websocket.remote_address[1]}'
        )
        print('[Path]', path)
    

    async def notify_all_user(self, message):
        '''
        If `self.users` is not empty, encrypt the
        `message`, decode it as unicode, and send it 
        to all `self.users`.
        '''
        if self.users:
            message = self.security.encrypt(message).decode()
            await asyncio.wait(
                [
                    user.send(message)
                    for user in self.users
                ]
            )
    

    def is_authorized(self, websocket):
        '''
        Validates the websocket authorization header 
        if its value is equal to `self.password`.
        
        It returns True if authorization header
        and `self.password` are equal. If websocket does
        not have authorization header or it can't be
        decrypted with `self.security`, it will return
        False.
        '''
        headers = self.get_headers(websocket)
        if not 'authorization' in headers:
            return False

        # Decrypt the websocket authorization header.
        websocket_password = self.security.decrypt(
            headers['authorization']
        )

        # If authorization header can't be decrypted,
        # return False.
        if websocket_password is None:
            return False

        if not websocket_password.decode() == self.password:
            return False
        return True
        

    def get_username_header(self, websocket):
        '''
        Encrypts the username from websocket header
        and return it. If not found, username is 
        an empty string, or username header can't be
        decrypted with `self.security`, it returns 
        None.
        '''
        headers = self.get_headers(websocket)
        if not 'username' in headers:
            return None

        # Decrypt the websocket username.
        websocket_username = self.security.decrypt(
            headers['username']
        )

        # If websocket username can't be decrypted
        # return None
        if websocket_username is None:
            return None

        if websocket_username.decode() == '':
            return None
        return websocket_username.decode()


    def get_headers(self, websocket):
        '''
        Returns the headers as dict from websocket.
        '''
        return dict([header for header in websocket.request_headers.raw_items()])
    

    def run(self):
        '''
        Start running the server.
        '''
        print(f'Server starts at `ws://localhost:{self.port}/`.')

        # Show the cryptography key of server to console
        # as client needs it for cryptography of authorization
        # and username headers, and messages of client.
        print('[Cryptography Key]')
        print('Please copy the following key below, you will need it for connecting to this server:')
        print(f'Key: {self.cryptography_key}')
        
        # Run the server forever.
        self.loop.run_until_complete(
            websockets.serve(self.server, 'localhost', self.port)
        )
        self.loop.run_forever()


from setuptools import setup

with open('README.md') as readmefile:
    long_description = readmefile.read()

setup(
    name='terminal-chatapp',
    version='1.0.0',
    author='Mark Carcillar',
    description='Secure communication on terminal.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/clediscover/terminal_chatapp',
    packages=['terminal_chatapp'],
    install_requires = [
        'websockets',
        'cryptography'
    ],
    python_requires='>=3.6'
)


import asyncio
import json
from random import random
from secrets import choice
from string import digits

import websockets

from .input import NonBlockingInput
from .event import create_message_event
from .security import Security


class Client:
    '''
    Client API for Terminal Chat Application.
    '''

    def __init__(
        self, 
        cryptography_key, 
        cryptography_digest_count=3, 
        url='ws://localhost:1719/', 
        username='', 
        password='top_secret'
        ):

        self.url = url
        self.loop = asyncio.get_event_loop()

        # For cryptography of messages of client, 
        # authorization and username header.
        self.security = Security(
            cryptography_key,
            cryptography_digest_count
        )

        # Set a message as an empty string since
        # this API is using a callback API for 
        # `input()` function.
        self.message = ''

        # Setup the credentials of client for connecting to server.
        self.username = username if username else self.create_username()
        self.headers = {
            'authorization': self.security.encrypt(password).decode(), 
            'username': self.security.encrypt(self.username).decode()
        }

        # Have an access to this Client API if connecting to the
        # server is successful. This should not be change by other
        # API.
        self._successfully_connected = False
    

    async def start_conversation(self):
        '''
        Real function that do the work for running the 
        whole `Client` API but you can use the `self.run()` 
        method for synchronous code.

        It connects to the websocket server (`self.url`)
        with an encrypted authorization and username header. 
        If the request to the server is rejected, this will 
        stop the program and show why client can't connect to 
        the server. Otherwise, if the request is accepted, 
        it will create a task for `self.chat_forever()` and 
        `self.receive_forever()` method and run them 
        concurrently. This is to allow the client to send 
        and receive a message at the same time.
        '''
        async with websockets.connect(
            self.url,
            extra_headers=self.headers
            ) as websocket:
            # Literally sleep for random milliseconds to ensure
            # that client request is accepted by the server. 
            # If not, it will just raise the 
            # `websockets.exceptions.ConnectionClosedError`
            # showing the status code and reason why client can't 
            # connect to the server.
            await self.sleepy_head()
            await websocket.ensure_open()

            # Show to console that connecting to server is successful
            # and set `self._successfully_connected` to True.
            self._successfully_connected = True
            print(f'Connected to `{self.url}` server.')
            print(f'You are connected as `{self.username}`.')

            # Run the `chat_forever()` and `receive_forever()`
            # concurrently.
            chat_task = asyncio.create_task(self.chat_forever(websocket))
            receive_task = asyncio.create_task(self.receive_forever(websocket))
            await chat_task
            await receive_task
        

    async def chat_forever(self, websocket):
        '''
        Have an input to the console so that,
        the client can type and send a message 
        to the websocket server. The message 
        is encrypted using the `self.security`
        before it sends to the websocket server.
        '''
        self.keyboard_thread = NonBlockingInput(self._set_message, '')
        while True:
            if not self.message == '':
                # Send the message as encrypted 
                # with `self.security`.
                self.message = self.security.encrypt(
                    create_message_event(self.username, self.message)
                ).decode()
                await websocket.send(self.message)
                self.message = ''
            await self.sleepy_head()
    

    async def receive_forever(self, websocket):
        '''
        Gets a message from the `websocket` server 
        every `self.sleepy_head()` and decrypts it
        using `self.security`.

        If event type is `users`, it shows how many 
        user is connected to the server. If event 
        type is `message`, it shows the message sender 
        username and its message content, but if the
        message sender is this client, it does not
        show the message.
        '''
        while True:
            rcv = await websocket.recv()
            rcv = json.loads(self.security.decrypt(rcv))
            if rcv['type'] == 'users':
                print('Users Connected:', rcv['users'])
            elif rcv['type'] == 'message':
                # Only show the message if it comes
                # from other websocket.
                if not self.username == rcv['from']:
                    print('[Receive]')
                    print('From:', rcv['from'])
                    print('Message:', rcv['message'])
            await self.sleepy_head()
    

    async def sleepy_head(self):
        '''
        Asyncio sleep for `random() ** 5`.
        '''
        await asyncio.sleep(random() ** 5)
    

    def create_username(self):
        '''
        Returns a random username started at `user_` 
        then random 4 numbers. For instance, `user_1234`.
        '''
        return 'user_' + ''.join(choice(digits) for _ in range(4))
    

    def _set_message(self, message):
        '''
        It set the `self.message` to `message` parameter.

        This method is used as callback on `NonBlockingInput` 
        from `self.chat_forever()` method and should not be 
        use by other API.
        '''
        self.message = message
    
    
    def run(self):
        '''
        Starts the client program.
        '''
        asyncio.run(self.start_conversation())

